# -*- coding: utf-8 -*-
import logging
from six import PY2, PY3
from peltak.actions import actions
from peltak.core import git, cstr, CommaSeparatedList
from peltak.common import prompt
from peltak.common.utils import extract_mr, mrdiff, resetmr_if_closed
from peltak.common.cache import CachedObj
from peltak.config import MR_CACHE_PATH
from peltak.system import forms
if PY2:
    from urlparse import urljoin
elif PY3:
    from urllib.parse import urljoin


@actions.register(
    'mr',
    desc=("Manage a Gitlab merge request. This will ask you all the details "
          "about the MR. The details can be "),

    cmdline=[{
        'args':     ('--ready',),
        'dest':     'ready',
        'action':   'store_true',
        'help': ("Mark merge request as ready. "
                 "This will remove [WIP] tag from the title")
    }, {
        'args':     ('-u', '--update'),
        'dest':     'force_update',
        'type':     CommaSeparatedList,
        'default':  [],
        'help': ("Comma separated list of fields to update in the description")
    }, {
        'args':     ('--skip-context',),
        'dest':     'skip_context',
        'action':   'store_true',
        'help': ("Skip filling template context")
    }]
)
def action_mr(app):
    app.authenticate()

    if resetmr_if_closed(app):
        print("Resetting cache, merge reqeust {} has been merged".format)

    mrcache = CachedObj.load_file(MR_CACHE_PATH)

    mr_details_form(
        api          = app.gitlab,
        mrcache      = mrcache,
        ready        = app.cl.ready,
        skip_context = app.cl.skip_context,
        force_update = app.cl.force_update
    )

    prompt.print_mr(app, mrcache)
    if 'branch_exists' in mrcache:
        need_push = not mrcache.branch_exists
    else:
        mrcache.branch_exists = mrcache.source_branch  in (
            b.name for b in app.gitlab.list_branches(app.conf.project_id)
        )
        need_push = not mrcache.branch_exists

    if need_push:
        print(cstr("^1Remote branch^0 ^33{}^0 ^1does not exist.^0".format(
            git.current_branch()
        )))
        msg = "Push to ^34{}^0?".format(app.conf.gitlab_url)
        if prompt.yesno(msg, default=False):
            git.push_current()
            mrcache.branch_exists = True

    if mrcache.branch_exists:
        if 'remote_cache' not in mrcache:
            msg = "\nCreate merge request on ^34{}^39".format(
                app.conf.gitlab_url
            )
            if prompt.yesno(cstr(msg), default=False):
                set_assignee(app.gitlab, mrcache)
                create_mr(app, mrcache)
                url = '/{}/merge_requests/{}'.format(
                    app.conf.project,
                    mrcache.remote_cache.iid
                )
                print(cstr("\nMerge request ^32#{}^0 created: ^34{}^39".format(
                    mrcache.remote_cache.iid,
                    urljoin(app.conf.gitlab_url, url)
                )))
        else:
            url = '/{}/merge_requests/{}'.format(
                app.conf.project,
                mrcache.remote_cache.iid
            )
            # Update instead of creating
            msg = "\nUpdate merge request ^32#{} ^34{}^39".format(
                mrcache.remote_cache.iid,
                urljoin(app.conf.gitlab_url, url)
            )
            if prompt.yesno(cstr(msg), default=True):
                set_assignee(app.gitlab, mrcache)
                update_mr(app, mrcache)


def set_assignee(api, mrcache):
    if 'assignee' not in mrcache:
        user = prompt.autouser(api, 'Assign to')
        if user is not None:
            mrcache.assignee = user


def mr_details_form(api, mrcache, ready, skip_context, force_update):
    mrcache.source_branch = str(git.current_branch())

    # Title
    if 'title' not in mrcache or 'title' in force_update:
        prefix = '' if ready else '[WIP] '
        mrcache.title = prefix + raw_input(cstr('^1Title:^0 '))
    else:
        if not ready and not mrcache.title.startswith('[WIP] '):
            mrcache.title = '[WIP] ' + mrcache.title
        elif ready and mrcache.title.startswith('[WIP] '):
            mrcache.title = mrcache.title[6:]
        prompt.form_print('Title', mrcache.title)

    # Assign To
    if 'assignee' in force_update:
        user = prompt.autouser(api, 'Assign to')
        if user is not None:
            mrcache.assignee = user
    elif 'assignee' in mrcache:
        prompt.form_print('Assign to', mrcache.assignee.username)

    # Description
    mrcache.description = forms.mr_description(
        mrcache,
        skip_context=skip_context,
        force_update=force_update
    )

    return mrcache


def create_mr(app, mrcache):
    mr = extract_mr(mrcache)
    logging.debug("Creating merge request: {}".format(mr))
    mrcache.remote_cache = app.gitlab.create_mr(app.conf.project_id, mr)


def update_mr(app, mrcache):
    diff = mrdiff(mrcache)
    if not diff:
        print(cstr("^1No changes, skipping^0"))
    else:
        logging.debug("Updating Merge request {} with: {}".format(
            mrcache.remote_cache.id, diff
        ))
        mrcache.remote_cache = app.gitlab.update_mr(
            app.conf.project_id, mrcache.remote_cache.id, diff
        )




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
import sys
import os
import os.path
import readline
import logging
import copy
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from peltak.config import ACTIONS_DIRS
from peltak.core import cstr
from peltak.system import forms
from peltak.config import LOG_FILENAME, MR_TEMPLATE_PATH
from peltak.actions import actions
from peltak.system.appcontext import AppContext
from peltak.common.utils import fetch_template


if not os.path.exists(os.path.dirname(LOG_FILENAME)):
    os.makedirs(os.path.dirname(LOG_FILENAME))
logging.basicConfig(filename=LOG_FILENAME, level=logging.DEBUG)


class ActionError(RuntimeError):
    pass



def mkcmdline():
    alist = sorted(actions.iter(), key=lambda x: x.name)
    cmdline = ArgumentParser(
        usage=cstr("^33%(prog)s^0 ^32action^0 [options] [action args]"),
        formatter_class = RawDescriptionHelpFormatter,
        description=cstr('\n'.join([
            "^33peltak^0 is a command line utility to improve your daily workflow.",
            "The main focus so far is making a gitlab merge request"
            " management easier.",
            "",
            "  List all available actions along with a short description",
            "  ^1$ %(prog)s list-actions^0",
            "",
            "  More detailed help about a particular action",
            "  ^1$ %(prog)s <command> --help^0",
            "",
            "Available actions:",
            "  " + '\n  '.join(cstr('^32{}^0').format(a.name) for a in alist)
        ]))
    )

    return cmdline


def preparse_cmdline():
    """ Extract action from the command line.

    Only the first argument can be an action. If it doesn't start with '-' it
    is considered one. Otherwise action = None will be returned along with the
    whole command line. If the action is found, this funtion will return
    the action name and the command line with action arg removed.
    """
    args = sys.argv[1:]
    action = None
    if len(args) > 0 and not args[0].startswith('-'):
        action = args[0]
        args = args[1:]
    return action, args


def parse_cmdline(cmdline):
    """ Parse command line and return action and args.

    This function assumes actions.discover() was already called. It will
    read the action name from the command line and find the matching action.
    If the action cannot be found, ``ActionError`` will be raised.

    It will collect all the command line arguments from the action and parse
    the command line using that specification.

    """
    aname, args = preparse_cmdline()
    action = actions.find(aname)

    if action is not None:
        if 'cmdline' in action.meta:
            for arg in action.meta.cmdline:
                kwargs = copy.copy(arg)
                kwargs.pop('args')
                cmdline.add_argument(*arg['args'], **kwargs)

        cmdline.description = ""
        cmdline.epilog = action.meta.get('desc', '')

    cl = cmdline.parse_args(args)
    if action is None:
        raise ActionError("'{}' action not found".format(aname))

    return action, cl


def main():
    actions.discover(ACTIONS_DIRS)
    readline.parse_and_bind('tab: complete')

    cl = mkcmdline()
    try:
        action, cl = parse_cmdline(cl)
        app = AppContext(cl)

        action.run(app)
    except ActionError as ex:
        print("\033[31m{}\033[0m".format(str(ex)))
        cl.print_help()


if __name__ == '__main__':
    main()




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
import subprocess


def current_branch():
    return subprocess.check_output(
        ['git', 'symbolic-ref', '--short', '-q', 'HEAD']
    ).strip()


def review_branch(name=None):
    """ use difftool to review given (current) branch.

    If no branch is passed, it will review the current branch against the
    master.
    """
    try:
        mybranch = current_branch()
        bspec = 'master..{}'.format(mybranch)
        subprocess.call(['git', 'difftool', '--tool=kompare', '-y', bspec])
    except subprocess.CalledProcessError:
        print("\033[31mFailed to run\033[0m \033[32mgit difftool\033[0m")


def push_current():
    """ Push current branch to master.

    This will call ``git push`` as a subprocess. All user creditentials are
    passed directly to git.
    """
    mybranch = current_branch()
    return subprocess.call(['git', 'push', 'origin', mybranch],
                           stdout=subprocess.PIPE)


def get_origin():
    """ Return the remote origin URL. """
    return subprocess.check_output(
        ['git', 'config', '--get', 'remote.origin.url']
    ).strip()


def get_cert():
    return subprocess.check_output(
        ['git', 'config', '--get', 'http.sslCaInfo']
    ).strip() or None




# -*- coding: utf-8 -*-
from peltak.actions import actions
from peltak.core import git, cstr
from peltak.config import MR_CACHE_PATH
from peltak.system import forms
from peltak.common.cache import CachedObj
from peltak.common.utils import resetmr_if_closed


@actions.register(
    'push',
    desc=('Push the current branch to origin. This will go throught the '
          'checklist if it exists in the template. Each question only has '
          'to be answered once. That answer is cached and for the moment '
          'cannot be changed (Feature planned for the future).')
)
def action_push(app):
    app.authenticate()

    resetmr_if_closed(app)
    mrcache = CachedObj.load_file(MR_CACHE_PATH)

    forms.mr_description(mrcache,
                         skip_context=True,
                         force_update=[])

    try:
        print(cstr("Pushing ^33{}^0 to ^34{}^0 ^1(^^C to skip)^0".format(
            git.current_branch(), app.conf.gitlab_url
        )))
        if git.push_current() == 0:
            mrcache.branch_exists = True
    except KeyboardInterrupt:
        pass




# -*- coding: utf-8 -*-
from peltak.actions import actions
from peltak.config import MR_CACHE_PATH
from peltak.common.cache import CachedObj


@actions.register(
    'mr-reset',
    desc=("Reset local merge request cache. This will erase the local copy of "
          "the current merge request.")
)
def resetmr(app):
    mrcache = CachedObj.load_file(MR_CACHE_PATH)
    mrcache.reset()




# -*- coding: utf-8 -*-
"""
Action manager.


**registering actions**


**Example action**

.. code-block:: python

    from peltak.actions import actions
    from peltak.core import git

    @actions.register('action-name')
    def myaction(app):
        print("Current branch: {}".format(git.current_branch()))
"""
from igor.plugins import SimplePluginManager
from ..config import ACTIONS_DIRS
actions = SimplePluginManager(ACTIONS_DIRS)



# -*- coding: utf-8 -*-
from igor.jsutil import json_dumps
from peltak.actions import actions
from peltak.common.cache import CachedObj
from peltak.config import MR_CACHE_PATH


@actions.register(
    'mr-get',
    desc="""
    Get remote merge request and save it in local cache. The description will
    probably have to be redone if you want to edit it using this app but at
    least the local cache will be connected to the selected repository. This
    is mostly useful after mr-reset  if you want to continue to work on a
    given merge request.
    """,
    cmdline=[{
        'args':     ('id',),
        'type':     int,
        'help':     "Merge request #ID"
    }]
)
def get_mr(app):
    app.authenticate()

    mr = app.gitlab.get_mr(app.conf.project_id, app.cl.id)
    print(json_dumps(mr))

    mrcache = CachedObj.load_file(MR_CACHE_PATH)
    mrcache.reset()
    mrcache.remote_cache = mr
    #mrcache = CachedObj.load_file(MR_CACHE_PATH)
    #mrcache.remote_cache = mr
    #mr = extract_mr(mrcache)
    #mrcache.remote_cache = app.gitlab.create_mr(app.conf.project_id, mr)
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import sys
import getpass
import os.path
from ..config import MR_TEMPLATE_PATH
from ..core import ensure_dirs
from ..common import prompt
from ..core.template import Template


def user_details():
    defuser = getpass.getuser()
    sys.stdout.write("User (default: {}): ".format(defuser))
    user = sys.stdin.readline().strip() or defuser
    password = getpass.getpass()
    return user, password


def login():
    user, pw = user_details()
    return user, pw, prompt.yesno("Keep logged in?", default=True)


def mr_description(mr, skip_context=False, force_update=[]):
    if not os.path.exists(MR_TEMPLATE_PATH):
        if 'description' in mr:
            return mr.description
        else:
            return prompt.ask("Description")

    ensure_dirs(MR_TEMPLATE_PATH)
    template = Template.load_from_file(MR_TEMPLATE_PATH)
    template.fill(mr)

    if not skip_context:
        skip = lambda k, v: v and k.lower() not in force_update
        filled = ((k, v) for k, v in template.context.items() if skip(k, v))
        missing = (k for k, v in template.context.items() if not skip(k, v))

        for key, value in filled:
            prompt.form_print(key, value)

        for key in missing:
            ans = prompt.ask(key, multiline=True)
            template.context[key] = ans
            if ans:
                mr[key] = ans

    missing = (k for k, v in template.checklist.items() if not v)
    for key in missing:
        ans = prompt.yesno(key, default=False)
        template.checklist[key] = ans
        if ans:
            mr[key] = ans

    return template.render()

